@@ -1,10 +1,12 @@
 # Copyright 2021-2024 Ping Data Intelligence
 import enum
 from datetime import datetime
-from typing import Any, Dict, List, NotRequired, Optional, TypedDict
+from typing import Any, NotRequired, TypedDict
+
+from ..common_types import PingMapsStatus
 
 
-class SOV_STATUS(str, enum.Enum):
+class SOV_STATUS(enum.StrEnum):
     PENDING = "PENDING"
     IN_PROGRESS = "IN_PROGRESS"
     ENRICHING = "ENRICHING"
@@ -21,9 +23,10 @@ INCOMPLETE_STATUSES = [
 ]
 
 
-class SOV_RESULT_STATUS(str, enum.Enum):
+class SOV_RESULT_STATUS(enum.StrEnum):
     SUCCESS = "SUCCESS"
     FAILED_TO_READ = "FAILED_TO_READ"
+    PARTIAL_PARSE = "PARTIAL_PARSE"
     FAILED_TO_PARSE = "FAILED_TO_PARSE"
     FAILED_TO_PROCESS = "FAILED_TO_PROCESS"
 
@@ -38,13 +41,15 @@ class FixSOVResponseRequest(TypedDict):
     pct_complete: int
 
 
-class OutputData(TypedDict):
-    label: str
-    # sovid: str
-    scrubbed_filename: str
-    # sov_data_id: int
-    output_format: str
+class FixSOVResponseResultInput(TypedDict):
+    """Results/details of processing of each of the input files."""
+
+    status: SOV_RESULT_STATUS
+    identified_document_type: str | None
     url: str
+    filename: str
+    # checksum: str
+    status_message: NotRequired[str | None]
 
 
 class FixSOVResponseResultOutput(TypedDict):
@@ -56,6 +61,7 @@ class FixSOVResponseResultOutput(TypedDict):
 class FixSOVResponseResult(TypedDict):
     message: str
     status: SOV_RESULT_STATUS
+    inputs: list[FixSOVResponseResultInput]
     outputs: list[FixSOVResponseResultOutput]
 
 
@@ -64,6 +70,14 @@ class FixSOVResponse(TypedDict):
     result: NotRequired[FixSOVResponseResult]
 
 
+class FixSOVProcessResponse(TypedDict):
+    success: bool
+    id: str
+    start_response: FixSOVResponse
+    final_response: FixSOVResponse
+    local_outputs: list[str] | None
+
+
 class UpdateOutputData(TypedDict):
     url: str
     output_format: str
@@ -77,11 +91,13 @@ class OutputData(TypedDict):
     # sov_data_id: int
     output_format: str
     url: str
+    ping_ready_id: NotRequired[str]
 
 
 class UpdateData(TypedDict):
     sovid: int
     sudid: int
+    status: str
     filename: str
     document_type: str
     record_type: str
@@ -91,6 +107,11 @@ class UpdateData(TypedDict):
     num_rows: int
     sov_data_last_updated_date: datetime
     outputs: list[UpdateOutputData]
+    output_formats: NotRequired[list[str]]
+    contract_terms_tracking_id: NotRequired[str]
+    contract_terms_policy_number: NotRequired[str]
+    completed_time: NotRequired[datetime]
+    client_ref: NotRequired[str | None]
 
 
 class SOVData(TypedDict):
@@ -102,6 +123,7 @@ class SOVData(TypedDict):
     status_display: str | None
     last_health_status: str | None
     client_ref: str | None
+    input_data: list[FixSOVResponseResultInput] | None
     output_data: list[OutputData] | None
     origin: str | None
     organization__short_name: str | None
@@ -109,7 +131,8 @@ class SOVData(TypedDict):
     ping_maps_url: str | None
     input_file_url: str | None
     extra_data: dict[str, Any] | None
-    updates: dict[int, UpdateData] | None
+    updates: dict[str, UpdateData] | None
+    all_updates: list[UpdateData] | None
     completed_time: str | None
     created_time: str | None
     subject: str | None
@@ -119,8 +142,9 @@ class SOVData(TypedDict):
     progress_started_time: str | None
     parsing_completed_time: str | None
     pingdata_stats: dict[str, dict[str, int]] | None
-    data_readiness_score: NotRequired[int]
-    data_readiness_notes: NotRequired[list[dict[str, str | int]]]
+    ping_maps: NotRequired[PingMapsStatus | None]
+    data_readiness_score: NotRequired[int | None]
+    data_readiness_notes: NotRequired[list[dict[str, str | int]] | None]
 
 
 class ActivityResponse(TypedDict):
